#include <string>
#include <iostream>
#include <cstdlib>
#include <complex>
#include <cmath>

#include "BufferedNode.h"
#include "Vector.h"
#include "Matrix.h"
#include "Map.h"
#include "Source.h"

#include "BF.h"
#include "BF_Matrix.h"
#include "BF_MSNR.h"
#include "ArrayInit.h"
#include "MICARY_struct.h"
#include <../config.h>
#include <iostream>

#ifdef ENABLE_HARKIO

using namespace std;
using namespace FD;

#ifdef ENABLE_HARKIO3
	#include "HarkFileIO_NH.h"
#elif  ENABLE_HARKIO2
	#include "HarkFileIO.h"
	using namespace harkio2;
#elif ENABLE_HARKIO1
	#include "HarkFileIO.h"
	using namespace harkio;
#endif

#include "HarkFileIOByMatrix.h"


class MSNR;

DECLARE_NODE(MSNR);
/*Node
 *
 * @name MSNR
 * @category HARK:Separation
 * @description Execute sound source separation with MSNR.
 *
 * @input_name INPUT_FRAMES
 * @input_type Matrix<complex<float> >
 * @input_description Input multi-channel spectrum. A row is a channel, and a column is a spectrum for the corresponding channel. 
 *
 * @input_name INPUT_SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description Separated spectrum with a source ID(key), and the value is the separated spectrum for each sound source (Vector<complex<float> >).
 *
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description The frame length of each channel (in samples) [default: 512].
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length beween adjacent frames (in samples)[default: 160].
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz) [default: 16000].
 * 
 * @parameter_name LOWER_BOUND_FREQUENCY
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Lower bound of frequency (Hz). [default: 0]
 * 
 * @parameter_name UPPER_BOUND_FREQUENCY
 * @parameter_type int
 * @parameter_value 8000
 * @parameter_description Upper bound of frequency (Hz). [default: 8000]
 *
 * @parameter_name DECOMPOSITION_ALGORITHM
 * @parameter_type string
 * @parameter_list GEVD:GSVD
 * @parameter_value GEVD
 * @parameter_description GEVD is more rubust but slow. GSVD is fast but less robust.
 *
 * @parameter_name ALPHA
 * @parameter_type float
 * @parameter_value 0.99
 * @parameter_description [NEW] ALPHA parameter used in MNSR as step size parameter.
 *
 * @parameter_name ENABLE_DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description enable debug print
 *
 END*/
#if 0
 * @input_name INPUT_NOISE_SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Noise source locations with ID. Each element of the vector is a source location with ID specified by "Source".
#endif

class MSNR : public BufferedNode
{
public:
    MSNR(string nodeName, ParameterSet params)
	: BufferedNode(nodeName, params)
    {
		/*== FlowDesigner 入出力インタフェース ==*/

		/* 入出力IDの割り当て */
		this->input_frameID        = addInput("INPUT_FRAMES");
		this->input_signalSourceID = addInput("INPUT_SOURCES");
#if 0
		this->input_noiseSourceID  = -1;//addInput("INPUT_NOISE_SOURCES");
#endif
		this->output_ID            = addOutput("OUTPUT");

		/*== パラメータの取得と検証 ==*/

		/* 周波数解析窓長パラメータ値の取得と検証 */
		this->length = dereference_cast<int>(parameters.get("LENGTH"));
		if (this->length <= 0) {
		    throw (new NodeException(NULL, "MSNR:LENGTH is invalid value.", __FILE__, __LINE__));
		}

		/* サンプリング周波数パラメータ値の取得と検証 */
		this->samplingRate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));
		if (this->samplingRate <= 0) {
		    throw (new NodeException(NULL, "MSNR:SAMPLING_RATE is invalid value.", __FILE__, __LINE__));
		}

		/* 上下限周波数の取得と検証 */
		int lower_bound_freq = dereference_cast<int>(parameters.get("LOWER_BOUND_FREQUENCY"));
		int upper_bound_freq = dereference_cast<int>(parameters.get("UPPER_BOUND_FREQUENCY"));
		if (lower_bound_freq < 0 || lower_bound_freq > this->samplingRate/2) {
	    	throw (new NodeException(NULL, "MSNR:LOWER_BOUND_FREQUENCY is invalid value.", __FILE__, __LINE__));
		}
		if (upper_bound_freq < 0 || upper_bound_freq > this->samplingRate/2) {
	    	throw (new NodeException(NULL, "MSNR:UPPER_BOUND_FREQUENCY is invalid value.", __FILE__, __LINE__));
		}
		if (lower_bound_freq > upper_bound_freq) {
	    	throw (new NodeException(NULL, "MSNR:LOWER_BOUND_FREQUENCY is larger than UPPER_BOUND_FREQUENCY.", __FILE__, __LINE__));
		}
		this->lowerfreq = floor((double)lower_bound_freq / (double)this->samplingRate * (double)this->length);
		this->upperfreq = ceil((double)upper_bound_freq / (double)this->samplingRate * (double)this->length);

		/* MSNR で用いる alpha 値 */
		if (parameters.exist("ALPHA")) {
		    this->alpha = dereference_cast<float>(parameters.get("ALPHA"));
		}
		if (parameters.exist("DECOMPOSITION_ALGORITHM")) {
		    this->decomposition_algorithm = object_cast<string>(parameters.get("DECOMPOSITION_ALGORITHM"));
		}
	    
		/* デバッグ出力用フラグ */
		if (parameters.exist("ENABLE_DEBUG")) {
			this->enable_debug = dereference_cast<bool>(parameters.get("ENABLE_DEBUG"));
		}

		/* adaptBF 作業変数確保 */
		this->bfAdapter = new BF::CMSNR (this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed, this->decomposition_algorithm, this->alpha);
	}
	
    virtual ~MSNR()
    {
		delete this->bfAdapter;
    }

public:
    virtual int
    translateInput(string inputName)
    {
		if (inputName == "INPUT_FRAMES") {
		    return this->input_frameID;
		} else if(inputName == "INPUT_SOURCES") {
		    return this->input_signalSourceID;
#if 0
		} else if(inputName == "INPUT_NOISE_SOURCES") {
		    return this->input_noiseSourceID = addInput("INPUT_NOISE_SOURCES");
#endif
		} else {
		    throw new NodeException(this, inputName + " is not supported.", __FILE__, __LINE__);
		}
    }

    void calculate(int output_id, int count, Buffer &out)
    {
		/*== 入出力の準備 ==*/

		/* スペクトル入力 */
		ObjectRef orInputTemp = getInput(this->input_frameID, count);
		Matrix<complex<float> > &inputValueMatrix = object_cast<Matrix<complex<float> > >(orInputTemp);
		
		int numMic  = inputValueMatrix.nrows();	// チャンネル数
		int numFreq = inputValueMatrix.ncols();	// 周波数ライン数
		
		this->bfAdapter->numMicChannels = numMic;
		    	    
		/* 信号音源位置情報入力 */
		RCPtr<Vector<ObjectRef> > pSignalSources = getInput(this->input_signalSourceID, count);
		const Vector<ObjectRef> *signalSources = pSignalSources.get();
		int numSource = signalSources->size();	// 信号音源数
#if 0
		int numSourceS = signalSources->size();	// 信号音源数
		    
		/* ノイズ音源位置情報入力 */
		RCPtr<Vector<ObjectRef> > pNoiseSources;
		const Vector<ObjectRef> *noiseSources = NULL;
		int numSourceN = 0;
		if (this->input_noiseSourceID >= 0) {
		    pNoiseSources = getInput(this->input_noiseSourceID, count);
		    noiseSources = pNoiseSources.get();
		    numSourceN = noiseSources->size(); // ノイズ音源数
		}

		int numSource  = numSourceS + numSourceN;	// 総音源数
//std::cout << "numSourceS: "<< numSourceS << ", numSourceN: " << numSourceN << std::endl;
#endif

		this->bfAdapter->numSignalSources = numSource; // add for MSNR

		/* 出力スペクトルの領域を準備 */
		ObjectRef orOutputTemp = ObjectRef(new Map<int,ObjectRef>);
		Map<int, ObjectRef> &mpOutputValue = object_cast<Map<int,ObjectRef> >(orOutputTemp);
		out[count] = orOutputTemp;
		mpOutputValue.clear();

		/*== 処理用データの準備 ==*/

		/* 入力スペクトル */
		BF::Matrix inputSpectrum(numMic, numFreq);
		for (int i = 0 ; i < numMic; i++) {
		    for (int j = 0; j < numFreq; j++) {
				complex<float> v = inputValueMatrix(i, j);
				inputSpectrum.setElementAt(i, j, BF::Complex(v.real(), v.imag()));
		    }
		}

		/* 出力スペクトル */
		BF::Matrix outputSpectrum(numSource, numFreq);

		/* 音源位置 */
		// MSNRは伝達関数ファイルを用いないため、src->compare_mode=Source::TFINDEX, DEGによる切り替えをしない
		std::vector<BF::SourceLocation> sourcePositions;
		for (int i = 0 ; i < numSource; i++) {
#if 0
		    RCPtr<Source> src = (i < numSourceS) ? signalSources->at(i) : noiseSources->at(i - numSourceS);
		    sourcePositions.push_back(BF::SourceLocation(src->x[0], src->x[1], src->x[2], src->tfindex[0], src->id, (i < numSourceS)));
#endif
		    RCPtr<Source> src = signalSources->at(i);
		    sourcePositions.push_back(BF::SourceLocation(src->x[0], src->x[1], src->x[2], src->tfindex[0], src->id, true));
		}

		if (numSource > 0) {
		    /*== 伝達関数計算（更新または継承）および分離行列初期化 ==*/
		    this->bfAdapter->updateSources(sourcePositions);

		    /*== MSNR 実行 ==*/
		    this->bfAdapter->executeBeamForming(inputSpectrum, outputSpectrum);

		// add↓
		}
		else { // numSource=0 の時
			// executeBeamForming　を通らないので、 adaptSeparationMatrices で使用する this->bfAdapter->inputSpectrum をここで更新
			inputSpectrum.swap(this->bfAdapter->inputSpectrum);
		}
		// add↑ : numSource=0 でも以下↓を実行する

		    /*== 分離行列適応処理 ==*/
		    this->bfAdapter->adaptSeparationMatrices();

		    /*== 処理結果を出力スペクトルの領域へ格納 ==*/
		    for (int i = 0 ; i < numSource; i++) {
				int id = sourcePositions[i].sourceID;
				ObjectRef orOutputTempV = ObjectRef(Vector<complex<float> >::alloc(numFreq));
				Vector<complex<float> > &vcfOutputValue = object_cast<Vector<complex<float> > > (orOutputTempV) ;
			    
				for (int j = 0 ; j < numFreq; j++) {
				    const BF::Complex v = outputSpectrum.getElementAt(i, j);
				    float re = v.real();
				    float im = v.imag();
				    if ((fpclassify(re) == FP_NAN || fpclassify(im) == FP_NAN) && (this->enable_debug)) {
						std::cout << "MSNR: Output Spectrum data ["<< i <<" , " <<  j << "] is NAN. Please check the parameters." << std::endl ;
				    }
				    if ((fpclassify(re) == FP_INFINITE || fpclassify(im) == FP_INFINITE) && (this->enable_debug)) {
						std::cout << "MSNR: Output Spectrum data ["<< i <<" , " <<  j << "] is INF. Please check the parameters." << std::endl ;
				    }
				    vcfOutputValue[j] = complex<float>(re, im);
				}
				mpOutputValue.insert(make_pair(id,orOutputTempV));
		    }

		// delete↓ : numSource=0 でも以上↑を実行する
		/*}*/
	}

#if 0 // HarkFileIOMatrixを使用するため
    void
    exportSeparationMatrices(const string				&filename,			/* 出力先ファイル名 */
			     int									numFreq,			/* 周波数ビンの数 */
			     int									samplingRate,		/* サンプリング周波数 */
			     const std::vector<BF::MicLocation>		&micPositions,		/* マイク位置 */
			     const std::vector<BF::SourceLocation>	&sourcePositions,	/* 音源位置 */
			     const std::vector<BF::Matrix>			&separationMatrices	/* 最新の分離行列群 */)
    {
		int numMic = micPositions.size();
		int numSrc = sourcePositions.size();

		MICARY_SMDatabase smDatabase = {0};
			
		SMDatabaseAlloc(&smDatabase, 0, 0, 0);
		SMDatabaseAlloc(&smDatabase, numFreq, numSrc, numMic);

		for (int i = 0; i < numMic; i++) {
		    const BF::MicLocation &ml = micPositions[i];
		    smDatabase.micary_locaMic.ha_idLocation[i] = i;
		    smDatabase.micary_locaMic.micary_Location[i].ha_fX = ml.x;
		    smDatabase.micary_locaMic.micary_Location[i].ha_fY = ml.y;
		    smDatabase.micary_locaMic.micary_Location[i].ha_fZ = ml.z;
		    smDatabase.micary_locaMic.micary_Location[i].ha_idPos = i;
		}

		for (int i = 0; i < numSrc; i++) {
		    const BF::SourceLocation &sl = sourcePositions[i];
		    smDatabase.micary_locaSource.ha_idLocation[i] = sl.sourceID;
		    smDatabase.micary_locaSource.micary_Location[i].ha_fX = sl.x;
		    smDatabase.micary_locaSource.micary_Location[i].ha_fY = sl.y;
		    smDatabase.micary_locaSource.micary_Location[i].ha_fZ = sl.z;
		    smDatabase.micary_locaSource.micary_Location[i].ha_idPos = sl.sourceID;
		}

		for (int i = 0 ; i < numFreq; i++) {
		    smDatabase.micary_smSeparation.micary_iaFreqLinex.ha_iValue[i] = (double)samplingRate * (i * 2 + 1) / 4 / numFreq;
		}

		int c = 0;
		for (int i = 0 ; i < numFreq; i++) {
		    const BF::Matrix &mat = separationMatrices[i];

		    for (int j = 0 ; j < numSrc; j++) {
				for (int k = 0 ; k < numMic; k++) {
				    smDatabase.micary_smSeparation.micary_cmpaSeparation.ha_cmpValue[c].re = mat.getElementAt(j, k).real();
				    smDatabase.micary_smSeparation.micary_cmpaSeparation.ha_cmpValue[c].im = mat.getElementAt(j, k).imag();
				    c++;
				}
		    }
		}

		MICARY_TFDBFileHeader tfdbfh;
		tfdbfh.lFs       = samplingRate;
		tfdbfh.lNFFT     = numFreq * 2;
		tfdbfh.lNMic     = numMic;
		tfdbfh.lNSrc     = numSrc;
		tfdbfh.lNFreq    = numFreq;
		tfdbfh.iVersion  = 1;
		tfdbfh.iFileType = FORMAT_TYPE_W;

		CHarkFileIO cHark;
		if (!cHark.WriteData(filename.c_str(), &tfdbfh,
				    smDatabase.micary_smSeparation.micary_cmpaSeparation,
				    smDatabase.micary_smSeparation.micary_iaFreqLinex,
				    smDatabase.micary_locaSource,
				    smDatabase.micary_locaMic,
				    "",
				    "",
				    "",
				    "",
				    "",
				    "GHDSS_EXPORT")) {
		    SMDatabaseFree(&smDatabase);
		    throw new NodeException(this, "Fail to Export file as HARK format.", __FILE__, __LINE__);
		}

		SMDatabaseFree(&smDatabase);
	}
#endif
    void
    exportSeparationMatrices(const string  &filename, /* 出力先ファイル名 */
							int samplingRate,		  /* サンプリング周波数 */
							int numFreq,			  /* 周波数ビンの数 */
    						BF::Database *SeparationMatricesDatabase)
    {
		CHarkFileIOByMatrix	cHarkMat;
		if (!cHarkMat.WriteWData(filename.c_str(), samplingRate, numFreq, *(SeparationMatricesDatabase))) {
		    throw new NodeException(this, "Fail to Export file as HARK format.", __FILE__, __LINE__);
		}
	}


private:
    /*=== FlowDesigner 入出力 ===*/

    /* フレームデータの入力ID */
    int input_frameID;

    /* 信号音源情報の入力ID */
    int input_signalSourceID;

#if 0
    /* ノイズ音源情報の入力ID */
    int input_noiseSourceID;
#endif

    /* 処理結果フレームデータの出力ID */
    int output_ID;

    /*=== パラメータ ===*/

    /* 周波数解析窓長 */
    int length;

    /* サンプリング周波数 */
    int samplingRate;
    
    /* 音速値 */
    float soundSpeed;
    
    /* 分析対象周波数 */
    int lowerfreq;
    int upperfreq;
    
    /* MSNR で用いる分解アルゴリズム */
    string decomposition_algorithm;

    /* MSNR で用いる alpha 値 */
    float alpha;
    
    /* デバッグ出力用フラグ */
    bool enable_debug;

    /*=== 作業用変数 ===*/


    /* 分離行列の更新を行う処理オブジェクト  */
    BF::BF *bfAdapter;

};

#endif
